(* 1. USAGE:

 * in ocaml interpreter
      ledit ocaml dynlink.cma camlp4o.cma extStream.cmo
      #use "<file>_parser.ml";;

 * Pour interpréter un fichier contenant "parser"
      ledit ocaml dynlink.cma camlp4o.cma 
      #use "<file>_parser.ml";;

 * Pour afficher la traduction en ocaml: 
      camlp4o -impl <file>_parser.ml

 * Pour générer  la traduction dans un fichier .ml:
      camlp4o -impl <file>_parser.ml -o <file>.ml

 * Pour interpréter le fichier ocaml généré: 
      ledit ocaml dynlink.cma camlp4o.cma extStream.cmo
      #use "<file>.ml";;
 *)
open Stream_ext
  
(* 2. How to parse a string with a stream parser ? *)
let (parse_with : (char Stream.t -> 'result) -> string -> (bool * 'result)) =
  fun my_parser string ->
    (print_string (String.concat "" [ "\nparser \""; string; "\" \n=\n" ]);
     Stream_ext.parse_with my_parser (Stream.of_string string))
  
(* 3. ==== Grammar of integers ====================

   Integer -1->  At_least_one_digit
   At_least_one_digit -2-> Digit . Some_Digit
   Some_Digit -3-> ""
   Some_Digit -4-> Digit . Some_Digit

  ================================================ *)
let rec (integer : char Stream.t -> int) =
  fun stream -> at_least_one_digit 0 stream
and (at_least_one_digit : int -> char Stream.t -> int) =
  fun i (__strm : _ Stream.t) ->
    let d = digit __strm
    in
      try some_digit ((10 * i) + d) __strm
      with | Stream.Failure -> raise (Stream.Error "")
and (digit : char Stream.t -> int) =
  fun (__strm : _ Stream.t) ->
    match Stream.peek __strm with
    | Some (('0' .. '9' as char)) ->
        (Stream.junk __strm; (int_of_char char) - (int_of_char '0'))
    | _ -> raise Stream.Failure
and (some_digit : int -> char Stream.t -> int) =
  fun i (__strm : _ Stream.t) ->
    match try Some (digit __strm) with | Stream.Failure -> None with
    | Some d ->
        (try some_digit ((10 * i) + d) __strm
         with | Stream.Failure -> raise (Stream.Error ""))
    | _ -> i
  
(* 4. ==== Grammar of identifiers ===================

   Ident  -> Letter Ident_Cont
   Ident_Cont -> "" 
   Ident_Cont -> (Letter | Digit | "_")  Ident_Cont
  =================================================== *)
let rec (ident : char Stream.t -> string) =
  fun (__strm : _ Stream.t) ->
    match Stream.peek __strm with
    | Some (('a' .. 'z' as char)) ->
        (Stream.junk __strm;
         let string =
           (try ident_cont __strm
            with | Stream.Failure -> raise (Stream.Error ""))
         in (String.make 1 char) ^ string)
    | _ -> raise Stream.Failure
and (ident_cont : char Stream.t -> string) =
  fun (__strm : _ Stream.t) ->
    match Stream.peek __strm with
    | Some (('a' .. 'z' | 'A' .. 'Z' | '0' .. '9' | '_' as char)) ->
        (Stream.junk __strm;
         let string =
           (try ident_cont __strm
            with | Stream.Failure -> raise (Stream.Error ""))
         in (String.make 1 char) ^ string)
    | _ -> ""
  
(* 5. === symbols ==== *)
let rec (symbol : char Stream.t -> string) =
  fun (__strm : _ Stream.t) ->
    match Stream.peek __strm with
    | Some
        (('(' | ')' | '[' | ']' | '{' | '}' | ',' | '!' | '%' | '@' | '*' |
            '+' | '-' | '/'
          as char))
        -> (Stream.junk __strm; String.make 1 char)
    | Some ((':' | ';' | '=' | '<' | '>' | '.' as char)) ->
        (Stream.junk __strm;
         let string =
           (try symbol_cont __strm
            with | Stream.Failure -> raise (Stream.Error ""))
         in (String.make 1 char) ^ string)
    | _ -> raise Stream.Failure
and (symbol_cont : char Stream.t -> string) =
  fun (__strm : _ Stream.t) ->
    match Stream.peek __strm with
    | Some ((':' | ';' | '=' | '<' | '>' | '.' as char)) ->
        (Stream.junk __strm;
         let string =
           (try symbol_cont __strm
            with | Stream.Failure -> raise (Stream.Error ""))
         in (String.make 1 char) ^ string)
    | _ -> ""
  
(* 6. How to transform a stream of char into a stream of token (= typed words) with space elimination *)
type keywords = string list

type symbols = string list

type token =
  | Int of int
  | Var of string
  | Kwd of string
  | Smb of string
  | Other of string

let (tokenize_with : keywords -> symbols -> char Stream.t -> token Stream.t) =
  fun keywords symbols stream ->
    let rec (tokenize : char Stream.t -> token Stream.t) =
      fun (__strm : _ Stream.t) ->
        match Stream.peek __strm with
        | Some (' ' | '\t' | '\n') -> (Stream.junk __strm; tokenize __strm)
        | _ ->
            (match try Some (integer __strm) with | Stream.Failure -> None
             with
             | Some int ->
                 let tokens =
                   (try tokenize __strm
                    with | Stream.Failure -> raise (Stream.Error ""))
                 in Stream.icons (Int int) tokens
             | _ ->
                 (match try Some (ident __strm) with | Stream.Failure -> None
                  with
                  | Some str ->
                      let tokens =
                        (try tokenize __strm
                         with | Stream.Failure -> raise (Stream.Error "")) in
                      let token =
                        if List.mem str keywords then Kwd str else Var str
                      in Stream.icons token tokens
                  | _ ->
                      (match try Some (symbol __strm)
                             with | Stream.Failure -> None
                       with
                       | Some str ->
                           let tokens =
                             (try tokenize __strm
                              with
                              | Stream.Failure -> raise (Stream.Error "")) in
                           let token =
                             if List.mem str symbols
                             then Smb str
                             else Other str
                           in Stream.icons token tokens
                       | _ -> Stream.sempty)))
    in tokenize stream
  
(* TEST *)
let (test_tokenize_with :
     keywords -> symbols -> string -> (bool * (token list))) =
  fun keywords symbols string ->
    let (bool, token_stream) =
      parse_with (tokenize_with keywords symbols) string
    in (bool, (Stream_ext.to_list token_stream))
  
let (test_tokenize : string -> (bool * (token list))) =
  test_tokenize_with [ "return"; "if"; "then"; "else"; "fi" ]
    [ ":="; ";"; "+"; "/"; "("; ")" ]
  

